#include "AYPlayer.h"
#include <memory.h>

#include "retrohack.h"
union RegPair
{
	Uint16 a;

#ifdef ALLEGRO_LITTLE_ENDIAN
	struct
	{
		Uint8 b0, b1;
	} b;
#else
	struct
	{
		Uint8 b1, b0;
	} b;
#endif
};

#define WriteToAF(r)\
	AReg = r.b.b1;\
	Misc = Sign = r.b.b0;\
	Zero = (r.b.b0&0x40)^0x40;\
	HalfCarry = r.b.b0&0x10;\
	Carry = r.b.b0&1;\
	LastOp = (r.b.b0&2) ? TRUE : FALSE;\
	Overflow = (r.b.b0&0x4) << 5;

union RegPair BC, DE, HL, SP, BC2, DE2, HL2, AF2, IX, IY, PC;
Uint8 AReg, RReg, RHigh;

Uint8 RReg, HalfCarry, Overflow, Zero, Carry, Misc, Sign, IReg, IMode;
unsigned char __Z80Mem[65536];
unsigned char ParityTab[256];
int LastOp, IFF1, IFF2;

#define ReadPort(a, b) b = ay_Read()
#define WritePort(a, b) ay_Write(a, b)

#define ReadFromAF(r)\
	r.b.b1 = AReg;\
	r.b.b0 = (Misc&0x28) | Carry | (LastOp ? 2 : 0) | ((Overflow >> 5)&0x4) | HalfCarry | (Zero ? 0 : 0x40) | (Sign&0x80);
	
#define Read8(n, v)	v = __Z80Mem[n.a];
#define Write8(n, v)	__Z80Mem[n.a] = v;

#define Push16(r)\
	SP.a--;\
	Write8(SP, r.b.b1);\
	SP.a--;\
	Write8(SP, r.b.b0);

#define Push16Instr(r)				Push16(r);

#define Pop16(r)\
	Read8(SP, r.b.b0);\
	SP.a++;\
	Read8(SP, r.b.b1);\
	SP.a++;

#define Dec16(r)\
	r.a--;

#define Inc16(r)\
	r.a++;

#define Or(r)\
	AReg |= r;\
	Sign = Zero = Misc = AReg; HalfCarry = Carry = 0; LastOp = FALSE;\
	Overflow = ParityTab[AReg];

#define Xor(r)\
	AReg ^= r;\
	Sign = Zero = Misc = AReg; HalfCarry = Carry = 0; LastOp = FALSE;\
	Overflow = ParityTab[AReg];

#define And(r)\
	AReg &= r;\
	Sign = Zero = Misc = AReg; HalfCarry = 0x10; Carry = 0; LastOp = FALSE;\
	Overflow = ParityTab[AReg];

#define Bit(r, v)\
	Sign = Misc = Zero = (r)&(v);\
	Overflow = Zero ? 0 : 0x80;\
	LastOp = FALSE;\
	HalfCarry = 0x10;

#define JR(cond)\
	Read8(PC, TempValue);\
	if(cond)\
	{\
		PC.a += (Sint8)TempValue;\
	}\
	PC.a++;

#define Ret(cond)\
	if(cond) {Pop16(PC);}

#define JP(cond)\
	Load16nn(Addr);\
	if(cond)	PC.a = Addr.a;

#define Call(cond)	\
	Load16nn(Addr);\
	if(cond)\
	{\
		SP.a--;\
		Write8(SP, PC.b.b1);\
		SP.a--;\
		Write8(SP, PC.b.b0);\
		PC.a = Addr.a;\
	}

#define Load1616(r1, r2)\
	r1.a = r2.a;

#define Load8n(r)\
	Read8(PC, r);\
	PC.a++;

#define Load16nn(r)\
	Read8(PC, r.b.b0); PC.a++;\
	Read8(PC, r.b.b1); PC.a++;

#define Load16nni(r)\
	Load16nn(Addr);\
	Read8(Addr, r.b.b0); Addr.a++;\
	Read8(Addr, r.b.b1);

#define Load8nn(r)\
	Load16nn(Addr);\
	Read8(Addr, r);

#define Loadnn16(r)\
	Load16nn(Addr);\
	Write8(Addr, r.b.b0); Addr.a++;\
	Write8(Addr, r.b.b1);

#define Inc8(r)\
	HalfCarry = ((r&0xf)+1)&0x10;\
	Overflow = (r == 127) ? 0x80 : 0;\
	r++;\
	LastOp = FALSE; Sign = Misc = Zero = r;

#define Dec8(r)\
	HalfCarry = ((r&0xf)-1)&0x10;\
	Overflow = (r == 128) ? 0x80 : 0;\
	r--;\
	LastOp = TRUE; Sign = Misc = Zero = r;

#define StHL()	Write8(HL, TempValue);

/*
	add - overflow occurs if two numbers of the same sign added yield a number with a
	different sign

	~AReg^n = 1 if numbers of same sign
	AReg^Temp16.b.b0 = 1 if result is of different sign

	sub - overflow occurs if the opposite is true

*/

#define Add16(r, n)\
	LastOp = FALSE;\
	Temp32 = r.a + n.a;\
	HalfCarry = (((r.a&0xfff) + (n.a&0xfff)) >> 8)&0x10;\
	Misc = Temp32 >> 8;\
	Carry = Temp32 >> 16;\
	r.a = Temp32;

#define Adc16(n)\
	LastOp = FALSE;\
	Temp32 = HL.a + n.a + Carry;\
	HalfCarry = (((HL.a&0xfff) + (n.a&0xfff) + Carry) >> 8)&0x10;\
	Sign = Misc = Temp32 >> 8;\
	Zero = Temp32 | Sign;\
	Carry = Temp32 >> 16;\
	Overflow = (~(HL.a^n.a)&(HL.a^Temp32)) >> 8;\
	HL.a = Temp32;

#define Sbc16(n)\
	LastOp = TRUE;\
	Temp32 = HL.a - n.a - Carry;\
	HalfCarry = (((HL.a&0xfff) - (n.a&0xfff) - Carry) >> 8)&0x10;\
	Misc = Sign = Temp32 >> 8;\
	Zero = Temp32 | Sign;\
	Carry = (Temp32 >> 16)&1;\
	Overflow = ((HL.a^n.a)&(HL.a^Temp32)) >> 8;\
	HL.a = Temp32;

#define Add8(n)\
	LastOp = FALSE;\
	Temp16.a = AReg + n;\
	HalfCarry = ((AReg&0xf) + (n&0xf))&0x10;\
	Carry = Temp16.b.b1;\
	Overflow = ~(AReg^n)&(AReg^Temp16.b.b0);\
	Misc = Sign = Zero = AReg = Temp16.b.b0;

#define Adc8(n)\
	LastOp = FALSE;\
	Temp16.a = AReg + n + Carry;\
	HalfCarry = ((AReg&0xf) + (n&0xf) + Carry)&0x10;\
	Carry = Temp16.b.b1;\
	Overflow = ~(AReg^n)&(AReg^Temp16.b.b0);\
	Misc = Sign = Zero = AReg = Temp16.b.b0;

#define Sbc8(n)\
	LastOp = TRUE;\
	Temp16.a = AReg - n - Carry;\
	HalfCarry = ((AReg&0xf) - (n&0xf) - Carry)&0x10;\
	Carry = Temp16.b.b1&1;\
	Overflow = (AReg^n)&(AReg^Temp16.b.b0);\
	Misc = Sign = Zero = AReg = Temp16.b.b0;

#define Sub8(n)\
	LastOp = TRUE;\
	Temp16.a = AReg - n;\
	HalfCarry = ((AReg&0xf) - (n&0xf))&0x10;\
	Carry = Temp16.b.b1&1;\
	Overflow = (AReg^n)&(AReg^Temp16.b.b0);\
	Misc = Sign = Zero = AReg = Temp16.b.b0;

#define Cp(n)\
	Temp32 = AReg - n;\
	Misc = n;\
	HalfCarry = ((AReg&0x0f) - (n&0x0f))&0x10;\
	Sign = Zero = Temp32;\
	Carry = (Temp32 >> 8)&1;\
	Overflow = (AReg^n)&(AReg^Temp32);\
	LastOp = TRUE;

#define Srl(n)\
	LastOp = FALSE; HalfCarry = 0;\
	Carry = n&1;\
	n >>= 1;\
	Overflow = ParityTab[Misc = Sign = Zero = n];

#define Sra(n)\
	LastOp = FALSE; HalfCarry = 0;\
	Carry = n&1;\
	n = (n >>1) | (n&0x80);\
	Overflow = ParityTab[Misc = Sign = Zero = n];

#define Rr(n)\
	LastOp = FALSE; HalfCarry = 0;\
	Temp8 = Carry;\
	Carry = n&1;\
	n = (n >> 1) | (Temp8 << 7);\
	Overflow = ParityTab[Misc = Sign = Zero = n];

#define Rrc(n)\
	LastOp = FALSE; HalfCarry = 0;\
	Carry = n&1;\
	n = (n >> 1) | (Carry << 7);\
	Overflow = ParityTab[Misc = Sign = Zero = n];

#define Sll(n)\
	LastOp = FALSE; HalfCarry = 0;\
	Carry = n >> 7;\
	n = (n << 1) | 1;\
	Overflow = ParityTab[Misc = Sign = Zero = n];

#define Sla(n)\
	LastOp = FALSE; HalfCarry = 0;\
	Carry = n >> 7;\
	n <<= 1;\
	Overflow = ParityTab[Misc = Sign = Zero = n];

#define Rl(n)\
	LastOp = FALSE; HalfCarry = 0;\
	Temp8 = Carry;\
	Carry = n >> 7;\
	n = (n << 1) | Temp8;\
	Overflow = ParityTab[Misc = Sign = Zero = n];

#define Rlc(n)\
	LastOp = FALSE; HalfCarry = 0;\
	Carry = n >> 7;\
	n = (n << 1) | Carry;\
	Overflow = ParityTab[Misc = Sign = Zero = n];

#define Res(r, v)	(r) &= ~(v);
#define Set(r, v)	(r) |= (v);

#define Out(port, value)\
	WritePort(port, value);

#define In(port, value)\
	ReadPort(port, value);\
	Overflow = ParityTab[Sign = Misc = Zero = value];\
	HalfCarry = 0;\
	LastOp = FALSE;

#define CPICompare()\
	Read8(HL, TempValue);\
\
	Temp16.a = AReg - TempValue;\
	HalfCarry = ((AReg&0x0f) - (TempValue&0x0f))&0x10;\
	Sign = Zero = Temp16.b.b0;\
	LastOp = TRUE;\
	if(HalfCarry) Temp16.a--;\
	Misc = ((Temp16.b.b0&0x01) << 4) | (Temp16.b.b0&0x08);\
\
	BC.a--;\
	Overflow = BC.a ? 0x80 : 0;

#define CopyHLToDE()\
	Read8(HL, TempValue);\
	Write8(DE, TempValue);\
\
	TempValue += AReg;\
	Misc = TempValue&0x8 | ((TempValue&0x2) << 4);\
	LastOp = FALSE;\
	HalfCarry = 0;\
\
	BC.a--;\
	Overflow = BC.a ? 0x80 : 0;

#define SetGroup(base, value)\
	case base:		Set(BReg, value);	break;\
	case base+1:	Set(CReg, value);	break;\
	case base+2:	Set(DReg, value)	break;\
	case base+3:	Set(EReg, value);	break;\
	case base+4:	Set(HReg, value);	break;\
	case base+5:	Set(LReg, value);	break;\
	case base+6:	Read8(HL, TempValue); Set(TempValue, value); StHL(); break;\
	case base+7:	Set(AReg, value);	break;

#define ResGroup(base, value)\
	case base:		Res(BReg, value);	break;\
	case base+1:	Res(CReg, value);	break;\
	case base+2:	Res(DReg, value);	break;\
	case base+3:	Res(EReg, value);	break;\
	case base+4:	Res(HReg, value);	break;\
	case base+5:	Res(LReg, value);	break;\
	case base+6:	Read8(HL, TempValue); Res(TempValue, value); StHL(); break;\
	case base+7:	Res(AReg, value);	break;

#define BitGroup(base, value)\
	case base:		Bit(BReg, value); break;\
	case base+1:	Bit(CReg, value); break;\
	case base+2:	Bit(DReg, value); break;\
	case base+3:	Bit(EReg, value); break;\
	case base+4:	Bit(HReg, value); break;\
	case base+5:	Bit(LReg, value); break;\
	case base+6:	Read8(HL, TempValue); Bit(TempValue, value); break;\
	case base+7:	Bit(AReg, value); break;

#define OpGroupReadNB(base, op)\
	case base:		op(BReg);			break;\
	case base+1:	op(CReg);			break;\
	case base+2:	op(DReg);			break;\
	case base+3:	op(EReg);			break;\
	case base+4:	op(HReg);			break;\
	case base+5:	op(LReg);			break;\
	case base+6:	Read8(HL, TempValue); op(TempValue); break;\
	case base+7:	op(AReg);			break;

#define BitwiseGroup(base, op)\
	op(base + 0x00, 0x01);\
	op(base + 0x08, 0x02);\
	op(base + 0x10, 0x04);\
	op(base + 0x18, 0x08);\
	op(base + 0x20, 0x10);\
	op(base + 0x28, 0x20);\
	op(base + 0x30, 0x40);\
	op(base + 0x38, 0x80);

#define OpGroupRead(base, op)\
	OpGroupReadNB(base, op)\
	case base+0x46:	Load8n(TempValue); op(TempValue);	break;

#define OpGroupModify(base, op, st)\
	case base:		op(BReg); break;\
	case base+st:	op(CReg); break;\
	case base+2*st:	op(DReg); break;\
	case base+3*st:	op(EReg); break;\
	case base+4*st:	op(HReg); break;\
	case base+5*st:	op(LReg); break;\
	case base+6*st:	Read8(HL, TempValue); op(TempValue);	StHL();	break;\
	case base+7*st:	op(AReg); break;

#define CondGroup(base, op)\
	case base+0x00:	op(NZ());	break;\
	case base+0x08:	op(Z());	break;\
	case base+0x10:	op(NC());	break;\
	case base+0x18:	op(C());	break;\
	case base+0x20:	op(PO());	break;\
	case base+0x28:	op(PE());	break;\
	case base+0x30:	op(P());	break;\
	case base+0x38:	op(M());	break;

#define BReg	BC.b.b1
#define CReg	BC.b.b0
#define DReg	DE.b.b1
#define EReg	DE.b.b0
#define HReg	HL.b.b1
#define LReg	HL.b.b0

#define NZ()	Zero
#define	Z()		!Zero
#define	NC()	!Carry
#define C()		Carry
#define PO()	!(Overflow&0x80)
#define PE()	(Overflow&0x80)
#define	P()		!(Sign&0x80)
#define M()		(Sign&0x80)

#define CalculateAddr(ip)\
	Read8(PC, Temp8); PC.a++;\
	Addr.a = ip.a + (Sint8)Temp8;

//#define GetT16()		Read8(Addr, TempValue);
#define PutT16()		Write8(Addr, TempValue);
#define GetT16()		Read8(Addr, TempValue);
#define IndexedOp(op)	op(TempValue); PutT16();
#define ResT16(b)		Write8(Addr, TempValue & ~b);
#define BitT16(b)		Bit(TempValue, b);
#define SetT16(b)		Write8(Addr, TempValue | b);

#define LdIP8(ip, v)\
	CalculateAddr(ip);\
	Write8(Addr, v);

#define Ld8IP(ip, v)\
	CalculateAddr(ip);\
	Read8(Addr, v);

#define Loadind(n)	Ld8IP(n, TempValue);

#define	IndexPage(ip)\
	Read8(PC, instr);\
	PC.a++;\
	RReg++;\
\
	switch(instr)\
	{\
	default: printf("fd %02x [%04x]\n", instr, PC.a); return;\
\
	case 0xcb :\
		CalculateAddr(ip);\
\
		Read8(PC, instr);\
		PC.a++;\
\
		GetT16();\
		switch(instr)\
		{\
			default: printf("fdcb d %02x\n", instr); return;\
\
			case 0x46: BitT16(0x01); break;\
			case 0x4e: BitT16(0x02); break;\
			case 0x56: BitT16(0x04); break;\
			case 0x5e: BitT16(0x08); break;\
			case 0x66: BitT16(0x10); break;\
			case 0x6e: BitT16(0x20); break;\
			case 0x76: BitT16(0x40); break;\
			case 0x7e: BitT16(0x80); break;\
\
			case 0x86: ResT16(0x01); break;\
			case 0x8e: ResT16(0x02); break;\
			case 0x96: ResT16(0x04); break;\
			case 0x9e: ResT16(0x08); break;\
			case 0xa6: ResT16(0x10); break;\
			case 0xae: ResT16(0x20); break;\
			case 0xb6: ResT16(0x40); break;\
			case 0xbe: ResT16(0x80); break;\
\
			case 0xc6: SetT16(0x01); break;\
			case 0xce: SetT16(0x02); break;\
			case 0xd6: SetT16(0x04); break;\
			case 0xde: SetT16(0x08); break;\
			case 0xe6: SetT16(0x10); break;\
			case 0xee: SetT16(0x20); break;\
			case 0xf6: SetT16(0x40); break;\
			case 0xfe: SetT16(0x80); break;\
\
			case 0x06: IndexedOp(Rlc);	break;\
			case 0x0e: IndexedOp(Rrc);	break;\
			case 0x16: IndexedOp(Rl);	break;\
			case 0x1e: IndexedOp(Rr);	break;\
			case 0x26: IndexedOp(Sla);	break;\
			case 0x2e: IndexedOp(Sra);	break;\
			case 0x36: IndexedOp(Sll);	break;\
			case 0x3e: IndexedOp(Srl);	break;\
		}\
		break;\
\
	case 0x24:	Inc8(ip.b.b1);		break;\
	case 0x25:	Dec8(ip.b.b1);		break;\
	case 0x26:	Load8n(ip.b.b1);	break;\
	case 0x84:	Add8(ip.b.b1);		break;\
	case 0x8c:	Adc8(ip.b.b1);		break;\
	case 0x94:	Sub8(ip.b.b1);		break;\
	case 0x9c:	Sbc8(ip.b.b1);		break;\
	case 0xa4:	And(ip.b.b1);		break;\
	case 0xac:	Xor(ip.b.b1);		break;\
	case 0xb4:	Or(ip.b.b1);		break;\
	case 0xbc:	Cp(ip.b.b1);		break;\
\
	case 0x44:	BReg = ip.b.b1;	break;\
	case 0x4c:	CReg = ip.b.b1; break;\
	case 0x54:	DReg = ip.b.b1; break;\
	case 0x5c:	EReg = ip.b.b1; break;\
	case 0x7c:	AReg = ip.b.b1; break;\
\
	case 0x60:	ip.b.b1 = BReg;	break;\
	case 0x61:	ip.b.b1 = CReg; break;\
	case 0x62:	ip.b.b1 = DReg; break;\
	case 0x63:	ip.b.b1 = EReg; break;\
	case 0x64:						break;\
	case 0x65:	ip.b.b1 = ip.b.b0; break;\
	case 0x67:	ip.b.b1 = AReg; break;\
\
	case 0x2c:	Inc8(ip.b.b0);		break;	/* inc ipl */\
	case 0x2d:	Dec8(ip.b.b0);		break;	/* dec ipl */\
	case 0x2e:	Load8n(ip.b.b0);	break;	/* ld ipl, n*/\
	case 0x85:	Add8(ip.b.b0);		break;	/* add ipl */\
	case 0x8d:	Adc8(ip.b.b0);		break;	/* adc ipl */\
	case 0x95:	Sub8(ip.b.b0);		break;	/* sub ipl */\
	case 0x9d:	Sbc8(ip.b.b0);		break;	/* sbc ipl */\
	case 0xa5:	And(ip.b.b0);		break;	/* and ipl */\
	case 0xad:	Xor(ip.b.b0);		break;	/* xor ipl */\
	case 0xb5:	Or(ip.b.b0);		break;	/* or ipl */\
	case 0xbd:	Cp(ip.b.b0);		break;	/* cp ipl */\
\
	case 0x45:	BReg = ip.b.b0; break;\
	case 0x4d:	CReg = ip.b.b0;	break;\
	case 0x55:	DReg = ip.b.b0;	break;\
	case 0x5d:	EReg = ip.b.b0;	break;\
	case 0x7d:	AReg = ip.b.b0;	break;\
\
	case 0x68:	ip.b.b0 = BReg;	break;\
	case 0x69:	ip.b.b0 = CReg;	break;\
	case 0x6a:	ip.b.b0 = DReg;	break;\
	case 0x6b:	ip.b.b0 = EReg;	break;\
	case 0x6c:	ip.b.b0 = ip.b.b1; break;\
	case 0x6d:						break;\
	case 0x6f:	ip.b.b0 = AReg;	break;\
\
	case 0x23:	Inc16(ip);			break;\
	case 0x2b:	Dec16(ip);			break;\
\
	case 0x34:	Loadind(ip); Inc8(TempValue); PutT16();	break;\
	case 0x35:	Loadind(ip); Dec8(TempValue); PutT16();	break;\
\
	case 0x09:	Add16(ip, BC);	break;\
	case 0x19:	Add16(ip, DE);	break;\
	case 0x29:	Add16(ip, ip);	break;\
	case 0x39:	Add16(ip, SP);	break;\
\
	case 0x21:	Load16nn(ip);	break;\
	case 0x22:	Loadnn16(ip);	break;\
	case 0x2a:	Load16nni(ip);	break;\
\
	case 0x86:	Loadind(ip); Add8(TempValue);	break;\
	case 0x8e:	Loadind(ip); Adc8(TempValue);	break;\
	case 0x96:	Loadind(ip); Sub8(TempValue);	break;\
	case 0x9e:	Loadind(ip); Sbc8(TempValue);	break;\
	case 0xa6:	Loadind(ip); And(TempValue);	break;\
	case 0xae:	Loadind(ip); Xor(TempValue);	break;\
	case 0xb6:	Loadind(ip); Or(TempValue);		break;\
	case 0xbe:	Loadind(ip); Cp(TempValue);		break;\
\
	case 0x70:	LdIP8(ip, BReg);	break;\
	case 0x71:	LdIP8(ip, CReg);	break;\
	case 0x72:	LdIP8(ip, DReg);	break;\
	case 0x73:	LdIP8(ip, EReg);	break;\
	case 0x74:	LdIP8(ip, HReg);	break;\
	case 0x75:	LdIP8(ip, LReg);	break;\
	case 0x77:	LdIP8(ip, AReg);	break;\
\
	case 0x46:	Ld8IP(ip, BReg);	break;\
	case 0x4e:	Ld8IP(ip, CReg);	break;\
	case 0x56:	Ld8IP(ip, DReg);	break;\
	case 0x5e:	Ld8IP(ip, EReg);	break;\
	case 0x66:	Ld8IP(ip, HReg);	break;\
	case 0x6e:	Ld8IP(ip, LReg);	break;\
	case 0x7e:	Ld8IP(ip, AReg);	break;\
\
	case 0x36:	\
		CalculateAddr(ip);\
		Read8(PC, TempValue); PC.a++;\
		PutT16();\
	break;	/* ld (ip+d), n */\
\
	case 0xe5:	Push16Instr(ip);		break;	/* push ip */\
	case 0xe1:	Pop16(ip);				break;	/* pop ip */\
	case 0xe3: /* ex (sp), ip */\
		Pop16(Temp16); Push16(ip);\
		ip.a = Temp16.a;\
	break;\
	case 0xe9:	PC.a = ip.a;	break;	/* jp (ip) */\
	case 0xf9:	Load1616(SP, ip);	break;	/* ld sp, ip */\
	}

#define LdGroup(base, r)\
	case base:		r = BReg;	break;\
	case base+1:	r = CReg;	break;\
	case base+2:	r = DReg;	break;\
	case base+3:	r = EReg;	break;\
	case base+4:	r = HReg;	break;\
	case base+5:	r = LReg;	break;\
	case base+6:	Read8(HL, r);		break;\
	case base+7:	r = AReg;	break;

#define LargeRegisterGroup(base, op)\
	case base:		op(BC);	break;\
	case base+0x10:	op(DE);	break;\
	case base+0x20:	op(HL);	break;\
	case base+0x30:	op(SP);	break;

#define WLargeRegisterGroup(base, op, ops)\
	case base:		op(BC, BBig, CBig);	break;\
	case base+0x10:	op(DE, DBig, EBig);	break;\
	case base+0x20:	op(HL, HBig, LBig);	break;\
	case base+0x30:	ops(SP);			break;

#define Add16HL(v) Add16(HL, v); 

void __Z80_Go()
{
	Uint8 LastInstr = 0;
	int MaxInstr = 2000;
	
	// do interrupt
	if(IFF2)
	{
		IFF2 = IFF1 = FALSE;
		Push16(PC);
		switch(IMode)
		{
			case 0:
			case 1:	PC.a = 0x0038;	break;
			case 2:
			{
				union RegPair Addr;
				Addr.b.b1 = IReg;
				Addr.b.b0 = 0xff;

				Read8(Addr, PC.b.b0); Addr.a++;
				Read8(Addr, PC.b.b1);
			}
			break;
		}
	}

	while(MaxInstr--)
	{
		union RegPair Addr;
		Uint8 Temp8, TempValue;
		union RegPair Temp16;
		Uint32 Temp32;
		Uint8 instr;

		/* fetch instruction */
		Read8(PC, instr); LastInstr = instr;
		PC.a++;
		RReg++;

		switch(instr)
		{
			default: printf("%02x [%04x]\n", instr, PC.a); return;

			case 0x00:					break; //NOP

			case 0x36:	
				Load8n(TempValue);
				StHL();
			break; //LD (HL), n

			LargeRegisterGroup(0x01, Load16nn);
			LargeRegisterGroup(0x03, Inc16);
			LargeRegisterGroup(0x09, Add16HL);
			LargeRegisterGroup(0x0b, Dec16);

			LdGroup(0x40, BReg);
			LdGroup(0x48, CReg);
			LdGroup(0x50, DReg);
			LdGroup(0x58, EReg);
			LdGroup(0x60, HReg);
			LdGroup(0x68, LReg);
			LdGroup(0x78, AReg);

			case 0x70:	Write8(HL, BReg);	break; //LD (HL), B
			case 0x71:	Write8(HL, CReg);	break; //LD (HL), C
			case 0x72:	Write8(HL, DReg);	break; //LD (HL), D
			case 0x73:	Write8(HL, EReg);	break; //LD (HL), E
			case 0x74:	Write8(HL, HReg);	break; //LD (HL), H
			case 0x75:	Write8(HL, LReg);	break; //LD (HL), L
			case 0x77:	Write8(HL, AReg);	break; //LD (HL), A

			case 0x06:	Load8n(BReg);		break; //LD B, n
			case 0x0e:	Load8n(CReg);		break; //LD C, n
			case 0x16:	Load8n(DReg);		break; //LD D, n
			case 0x1e:	Load8n(EReg);		break; //LD E, n
			case 0x26:	Load8n(HReg);		break; //LD H, n
			case 0x2e:	Load8n(LReg);		break; //LD L, n
			case 0x3e:	Load8n(AReg);		break; //LD A, n

			case 0x0a:	Read8(BC, AReg);	break; //LD A, (BC)
			case 0x02:	Write8(BC, AReg);	break; //LD (BC), A
			case 0x1a:	Read8(DE, AReg);	break; //LD A, (DE)
			case 0x12:	Write8(DE, AReg);	break; //LD (DE), A

			OpGroupRead(0x80, Add8);
			OpGroupRead(0x88, Adc8);
			OpGroupRead(0x90, Sub8);
			OpGroupRead(0x98, Sbc8);
			OpGroupRead(0xa0, And);
			OpGroupRead(0xa8, Xor);
			OpGroupRead(0xb0, Or);
			OpGroupRead(0xb8, Cp);

			OpGroupModify(0x05, Dec8, 8);
			OpGroupModify(0x04, Inc8, 8);

			case 0x10:	BReg--; JR(BReg); break; //DJNZ

			case 0x18:	JR(1);		break; //JR
			case 0x20:	JR(NZ());	break; //JR NZ
			case 0x28:	JR(Z());	break; //JR Z
			case 0x30:	JR(NC());	break; //JR NC
			case 0x38:	JR(C());	break; //JR C

			CondGroup(0xc0, Ret);
			CondGroup(0xc2, JP);
			CondGroup(0xc4, Call);

			case 0xc3:	JP(1);		break; //JP
			case 0xc9:	Pop16(PC);	break; //RET
			case 0xcd:	Call(1);	break; //CALL

			case 0xf3:	IFF1 = IFF2 = FALSE;	break; //DI
			case 0xfb:	IFF1 = IFF2 = TRUE;		break; //EI

			case 0x22:	Loadnn16(HL);		break; //LD (nn), HL
			case 0x32:
				Load16nn(Addr);
				Write8(Addr, AReg);
			break; //LD (nn), A
			case 0x2a:	Load16nni(HL);		break; //LD HL, (nn)
			case 0x3a:	Load8nn(AReg);	break; //LD A, (nn)

			case 0xf9:	Load1616(SP, HL);	break; //LD SP, HL

			case 0xc5:	Push16Instr(BC);			break; //PUSH BC
			case 0xd5:	Push16Instr(DE);			break; //PUSH DE
			case 0xe5:	Push16Instr(HL);			break; //PUSH HL
			case 0xf5:	ReadFromAF(Temp16); Push16Instr(Temp16);	break; //PUSH AF

			case 0xc1:	Pop16(BC);				break; //POP BC
			case 0xd1:	Pop16(DE);				break; //POP DE
			case 0xe1:	Pop16(HL);				break; //POP HL
			case 0xf1:	Pop16(Temp16); WriteToAF(Temp16); 	break; //POP AF

			case 0xd3:
				Load8n(Addr.b.b0);
				Addr.b.b1 = AReg;
				WritePort(Addr.a, AReg);
			break; //OUT (n), A

			case 0xdb:
				Load8n(Addr.b.b0);
				Addr.b.b1 = AReg;
				ReadPort(Addr.a, AReg);
			break; //IN A, (n)

			case 0xd9:
				BC.a ^= BC2.a; BC2.a ^= BC.a; BC.a ^= BC2.a;
				HL.a ^= HL2.a; HL2.a ^= HL.a; HL.a ^= HL2.a;
				DE.a ^= DE2.a; DE2.a ^= DE.a; DE.a ^= DE2.a;
			break; //EXX

			case 0xeb:
				DE.a ^= HL.a; HL.a ^= DE.a; DE.a ^= HL.a;
			break; //EX DE, HL

			case 0xe3:
				Pop16(Temp16); Push16(HL);
				
				HL.a = Temp16.a;
			break; //EX (SP), HL

			case 0x08:
				ReadFromAF(Temp16);
				WriteToAF(AF2);
				AF2.a = Temp16.a;
			break; //EX AF, AF'

			case 0xe9:	PC.a = HL.a;	break; //JP (HL)

			case 0x0f:
				Carry = AReg&1;
				Misc = AReg = (AReg >> 1) | (Carry << 7);
				LastOp = FALSE;
				HalfCarry = 0;
			break; //RRCA

			case 0x1f:
				Temp8 = Carry;
				Carry = AReg&1;
				Misc = AReg = (AReg >> 1) | (Temp8 << 7);
				LastOp = FALSE;
				HalfCarry = 0;
			break; //RRA

			case 0x07:
				Carry = AReg >> 7;
				Misc = AReg = (AReg << 1) | Carry;
				LastOp = FALSE;
				HalfCarry = 0;
			break; //RLCA

			case 0x17:
				Temp8 = Carry;
				Carry = AReg >> 7;
				Misc = AReg = (AReg << 1) | Temp8;
				LastOp = FALSE;
				HalfCarry = 0;
			break; //RLA

			case 0xc7:	Push16(PC);	PC.a = 0x00;	break; //RST 00h
			case 0xcf:	Push16(PC);	PC.a = 0x08;	break; //RST 08h
			case 0xd7:	Push16(PC);	PC.a = 0x10;	break; //RST 10h
			case 0xdf:	Push16(PC);	PC.a = 0x18;	break; //RST 18h
			case 0xe7:	Push16(PC);	PC.a = 0x20;	break; //RST 20h
			case 0xef:	Push16(PC);	PC.a = 0x28;	break; //RST 28h
			case 0xf7:	Push16(PC);	PC.a = 0x30;	break; //RST 30h
			case 0xff:	Push16(PC);	PC.a = 0x38;	break; //RST 38h

			case 0x2f:
				AReg ^= 0xff;
				Misc = AReg;
				HalfCarry = 0x10;
				LastOp = TRUE;
			break; //CPL

			case 0x27:
				TempValue =
					((HalfCarry || ((AReg&0xf) > 9)) ? 0x06 : 0x00) |
					((Carry || (AReg > 0x9f)) ? 0x60 : 0x00);

				if(LastOp)
				{
					/* sub8 without overflow calculation, and with carry adjustment */
					Temp16.a = AReg - TempValue;
					HalfCarry = ((AReg&0xf) - (TempValue&0xf))&0x10;
					Carry |= Temp16.b.b1&1;
					Misc = Sign = Zero = AReg = Temp16.b.b0;
				}
				else
				{
					if((AReg > 0x90) && ((AReg&0xf) > 9)) TempValue |= 0x60;

					/* add8 without overflow calculation, and with carry adjustment */
					Temp16.a = AReg + TempValue;
					HalfCarry = ((AReg&0xf) + (TempValue&0xf))&0x10;
					Carry |= Temp16.b.b1;
					Misc = Sign = Zero = AReg = Temp16.b.b0;
				}

				Overflow = ParityTab[AReg];
			break; //DAA

			case 0x76: return; //HALT

			case 0x37: Misc = AReg; Carry = 1; LastOp = FALSE; HalfCarry = 0; break; //SCF
			case 0x3f:
				Misc = AReg; LastOp = FALSE;
				HalfCarry = Carry << 4; Carry ^= 1;
			break; //CCF

			case 0xdd:	IndexPage(IX);	break; /* DD */
			case 0xfd:	IndexPage(IY);	break; /* FD */

			/* CB */
			case 0xcb:
				Read8(PC, instr);
				PC.a++;
				RReg++;

				switch(instr)
				{
					OpGroupModify(0x00, Rlc, 1);
					OpGroupModify(0x08, Rrc, 1);

					OpGroupModify(0x10, Rl, 1);
					OpGroupModify(0x18, Rr, 1);

					OpGroupModify(0x20, Sla, 1);
					OpGroupModify(0x28, Sra, 1);

					OpGroupModify(0x30, Sll, 1);
					OpGroupModify(0x38, Srl, 1);

					BitwiseGroup(0x40, BitGroup);
					BitwiseGroup(0x80, ResGroup);
					BitwiseGroup(0xc0, SetGroup);
				}
			break;

			/* ED */
			case 0xed:
				Read8(PC, instr);
				PC.a++;
				RReg++;

				switch(instr)
				{
					default: printf("ed %02x [%04x]\n", instr, PC.a); return;

					case 0x46: case 0x4e: case 0x66: case 0x6e: IMode = 0; break; //IM 0
					case 0x56: case 0x76: IMode = 1; break; //IM 1
					case 0x5e: case 0x7e: IMode = 2; break; //IM 2

					case 0x41: Out(BC.a, BReg);	break; //OUT (C), B
					case 0x49: Out(BC.a, CReg);	break; //OUT (C), C
					case 0x51: Out(BC.a, DReg);	break; //OUT (C), D
					case 0x59: Out(BC.a, EReg);	break; //OUT (C), E
					case 0x61: Out(BC.a, HReg);	break; //OUT (C), H
					case 0x69: Out(BC.a, LReg);	break; //OUT (C), L
					case 0x79: Out(BC.a, AReg);	break; //OUT (C), A

					case 0x71: Out(BC.a, 0);	break; //OUT (C), 0

					case 0x40: In(BC.a, BReg);	break; //IN B, (C)
					case 0x48: In(BC.a, CReg);	break; //IN C, (C)
					case 0x50: In(BC.a, DReg);	break; //IN D, (C)
					case 0x58: In(BC.a, EReg);	break; //IN E, (C)
					case 0x60: In(BC.a, HReg);	break; //IN H, (C)
					case 0x68: In(BC.a, LReg);	break; //IN L, (C)
					case 0x78: In(BC.a, AReg);	break; //IN A, (C)

					case 0x70: In(BC.a, Temp8);	break; //IN F, (C)

					case 0x44: case 0x4c: case 0x54: case 0x5c:
					case 0x64: case 0x6c: case 0x74: case 0x7c:
						TempValue = AReg;
						AReg = 0;
						Sub8(TempValue);
					break; //NEG

					case 0x47:	IReg = AReg; break; //LD I, A
					case 0x57:
						Misc = Sign = Zero = AReg = IReg;
						LastOp = FALSE;
						Overflow = IFF2 ? 0x80 : 0;
						HalfCarry = 0;
					break; //LD A, I

					case 0x4f:
						RReg = AReg;
						RHigh = AReg&0x80;
					break; //LD R, A
					case 0x5f:
						Misc = Sign = Zero = AReg = (RReg&0x7f) | RHigh;
						LastOp = FALSE;
						Overflow = IFF2 ? 0x80 : 0;
						HalfCarry = 0;
					break; //LD A, R

					LargeRegisterGroup(0x42, Sbc16);
					LargeRegisterGroup(0x43, Loadnn16);
					LargeRegisterGroup(0x4a, Adc16);
					LargeRegisterGroup(0x4b, Load16nni);

					case 0xb8:
						CopyHLToDE();
						DE.a--;	HL.a--;
						if(BC.a) {PC.a -= 2; }
					break; //LDDR

					case 0xb0:
						CopyHLToDE();
						DE.a++;	HL.a++;
						if(BC.a) {PC.a -= 2; }
					break; //LDIR

					case 0xa0:
						CopyHLToDE();
						DE.a++;	HL.a++;
					break; //LDI

					case 0xa8:
						CopyHLToDE();
						DE.a--;	HL.a--;
					break; //LDD


					case 0xa1: CPICompare(); HL.a++; break; //CPI
					case 0xa9: CPICompare(); HL.a--; break; //CPD

					case 0xb1:
						CPICompare();
						HL.a++;
						if(BC.a && Zero) {PC.a -= 2; }
					break; //CPIR

					case 0xb9:
						CPICompare();
						HL.a--;
						if(BC.a && Zero) {PC.a -= 2; }
					break; //CPDR


					case 0xab:
						Read8(HL, TempValue); HL.a--;
						Out(BC.a, TempValue);
						Dec8(BReg);
					break; //OUTD

					case 0xa3:
						Read8(HL, TempValue); HL.a++;
						Out(BC.a, TempValue);
						Inc8(BReg);
					break; //OUTI


					case 0x6f:
						Read8(HL, Temp16.b.b0);

						TempValue = (Temp16.b.b0 << 4) | (AReg&0x0f);
						AReg = (AReg&0xf0) | (Temp16.b.b0 >> 4);

						StHL();

						Overflow = ParityTab[Sign = Zero = Misc = AReg];
						LastOp = FALSE; HalfCarry = 0;
					break; //RLD

					case 0x67:
						Read8(HL, Temp16.b.b0);

						TempValue = (Temp16.b.b0 >> 4) | (AReg << 4);
						AReg = (AReg&0xf0) | (Temp16.b.b0&0x0f);

						StHL();

						Overflow = ParityTab[Sign = Zero = Misc = AReg];
						LastOp = FALSE; HalfCarry = 0;
					break; //RRD

					case 0x45: case 0x4d: case 0x55: case 0x5d:
					case 0x65: case 0x6d: case 0x75: case 0x7d:
						IFF1 = IFF2;
						Ret(1);
					break; //RETN
				}
			break;
		}
	}
}

/*
	Rant to make myself feel better:
	
		The AY file format is possibly the worst file format with the least helpful documentation that
		I have ever tried to deal with. This truly is purgatory.

*/
void __ayf_InstallCode(unsigned short Init, unsigned short Interrupt)
{
	__Z80Mem[0] = 0xf3; //DI
	__Z80Mem[1] = 0xcd; //CALL INIT
	__Z80Mem[2] = Init&0xff;
	__Z80Mem[3] = Init >> 8;
	if(Interrupt)
	{
		__Z80Mem[4] = 0xed; //LOOP: IM 1
		__Z80Mem[5] = 0x56;
		__Z80Mem[6] = 0xfb; //EI
		__Z80Mem[7] = 0x76; //HALT
		__Z80Mem[8] = 0xcd; //CALL INTERRUPT
		__Z80Mem[9] = Interrupt&0xff;
		__Z80Mem[10] = Interrupt >> 8;
		__Z80Mem[11] = 0x18; //JR LOOP
		__Z80Mem[12] = 4-13;
	}
	else
	{
		__Z80Mem[4] = 0xed; //LOOP: IM 1
		__Z80Mem[5] = 0x5e;
		__Z80Mem[6] = 0xfb; //EI
		__Z80Mem[7] = 0x76; //HALT
		__Z80Mem[8] = 0x18; //JR LOOP
		__Z80Mem[9] = 4-10;
	}
}

int __ayf_HasFile = 0;

int ayf_OpenFile(const char *name, int songnum)
{
	unsigned char PlayerVersion, NumSongs, FirstSong, HiReg, LoReg;
	int c;
	unsigned short SongOff, DataOff, InitAddr, InterruptAddr;
	int PointsPos, BlocksPos;
	FILE *input;
	char Buf[5];
	if(!(input = fopen(name, "rb")))
		return FALSE;

	Buf[4] = '\0';
	fread(Buf, 1, 4, input);
	if(strcmp(Buf, "ZXAY"))
	{
		fclose(input);
		return FALSE;
	}
	fread(Buf, 1, 4, input);
	if(strcmp(Buf, "EMUL"))
	{
		fclose(input);
		return FALSE;
	}

	// prepare memory
	memset(__Z80Mem, 0xff, 0x4000);
	memset(__Z80Mem, 0xc9, 0x100);
	memset(&__Z80Mem[0x4000], 0x00, 0x4000);
	__Z80Mem[0x38] = 0xfb;

	// read player version (why?)
	fseek(input, 1, SEEK_CUR);	// skip song file version
	PlayerVersion = fgetc(input);
	fseek(input, 2, SEEK_CUR); // skip specialplayer
	fseek(input, 2, SEEK_CUR); // skip author
	fseek(input, 2, SEEK_CUR); // skip misc

	NumSongs = fgetc(input);
	if(songnum > NumSongs)
	{
		fclose(input);
		return FALSE;
	}
	FirstSong = fgetc(input);

	// song offset. A completely useless feature of this file format
	SongOff = fgetc(input) << 8;
	SongOff |= fgetc(input);
	fseek(input, SongOff-2+ (4*songnum), SEEK_CUR);

	fseek(input, 2, SEEK_CUR); // offset to song name. ARGH! Why so many useless tables, AY?
	DataOff = fgetc(input) << 8;
	DataOff |= fgetc(input);
	fseek(input, DataOff-2, SEEK_CUR); // seek to song data. This file format is horrendous

	fseek(input, 8, SEEK_CUR); // skip stupid Amiga proprietary junk. Note: AY docs incorrectly list SongLength and FadeLength as words
	HiReg = fgetc(input);
	LoReg = fgetc(input);

	// get location of points data
	PointsPos = fgetc(input) << 8;
	PointsPos |= fgetc(input);
	PointsPos += ftell(input)-2;

	// get location of addresses data
	BlocksPos = fgetc(input) << 8;
	BlocksPos |= fgetc(input);
	BlocksPos += ftell(input)-2;

	//get points first. Another completely unnecessary file jump
	fseek(input, PointsPos, SEEK_SET);
	SP.b.b1 = fgetc(input);
	SP.b.b0 = fgetc(input);
	InitAddr = fgetc(input) << 8;
	InitAddr |= fgetc(input);
	InterruptAddr = fgetc(input) << 8;
	InterruptAddr |= fgetc(input);

	// do some more Z80 stuffing, accordingly
	if(InitAddr) __ayf_InstallCode(InitAddr, InterruptAddr);

	//stuff Z80 memory
	fseek(input, BlocksPos, SEEK_SET);
	while(1)
	{
		unsigned short Addr, Length, Offset;	//ignore AY docs: NOT words!
		int curpos;
		Addr = fgetc(input) << 8;
		Addr |= fgetc(input);
		if(!Addr) 
			break;
		if(!InitAddr)
		{
			__ayf_InstallCode(InitAddr = Addr, InterruptAddr);
		}
		Length = fgetc(input) << 8;
		Length |= fgetc(input);
		Offset = fgetc(input) << 8;
		Offset |= fgetc(input);

		curpos = ftell(input);
		fseek(input, Offset-2, SEEK_CUR);
		fread(&__Z80Mem[Addr], 1, Length, input);
		fseek(input, curpos, SEEK_SET);
	}
	fclose(input);

	c = 256;
	while(c--)
	{
		ParityTab[c] = c^1;
		ParityTab[c] = (ParityTab[c] >> 4) ^ (ParityTab[c] & 15);
		ParityTab[c] = (ParityTab[c] >> 2) ^ (ParityTab[c] & 3);
		ParityTab[c] = (ParityTab[c] >> 1) ^ (ParityTab[c] & 1);
		ParityTab[c] <<= 7;
	}

	PC.a = 0;
	AF2.b.b1 = HL.b.b1 = HL2.b.b1 = DE.b.b1 = DE2.b.b1 = BC.b.b1 = BC2.b.b1 = IX.b.b1 = IY.b.b1 = HiReg;
	AF2.b.b0 = HL.b.b0 = HL2.b.b0 = DE.b.b0 = DE2.b.b0 = BC.b.b0 = BC2.b.b0 = IX.b.b0 = IY.b.b0 = LoReg;
	WriteToAF(AF2);
	IFF1 = IFF2 = FALSE;
	IMode = 0;
	IReg = 3;
	Carry = 0;

	return __ayf_HasFile = TRUE;
}

void ayf_CloseFile(void)
{
	__ayf_HasFile = FALSE;
}

int ayf_Update(void)
{
	// [if a song is loaded] go Z80!
	if(__ayf_HasFile)
	{
		__Z80_Go();
		return TRUE;
	}
	return FALSE;
}
